/* rockywd.c copyright (c) 2000 by Crusader Systems (Pty) Ltd www.crusader.co.za
 * author: Charl P. Botha <cpbotha@ieee.org>
 * $Id: rockywd.c,v 1.3 2000/03/15 13:18:36 cpbotha Exp $
 * 
 * linux driver for Rocky-3702EV SBC built-in watchdog
 * http://www.icpacquire.com.tw/3702EV.htm
 * 
 * thanks go to Crusader Systems (Pty) Ltd who have graciously allowed 
 * publishing under GPL
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Charl P. Botha <cpbotha@ieee.org> http://cpbotha.net
 * 
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/miscdevice.h>
#include <linux/watchdog.h>
#include <linux/malloc.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <linux/notifier.h>
#include <linux/reboot.h>
#include <linux/init.h>


#define WDIO 0x443
#define WDIODIS 0x43

static int rockywd_is_open=0;

static int rockywd_open(struct inode *inode, struct file *file) {
   if (MINOR(inode->i_rdev)!=WATCHDOG_MINOR) return -ENODEV;
   /* make sure of exclusive access */
   if (rockywd_is_open) return -EBUSY;
   MOD_INC_USE_COUNT;
   rockywd_is_open=1;

   /* enable the watchdog by reading from this port; complicated huh? */
   inb_p(WDIO);
   
   return 0;
}

static int rockywd_release(struct inode *inode, struct file *file) {
   if (MINOR(inode->i_rdev)==WATCHDOG_MINOR) {
#ifndef CONFIG_WATCHDOG_NOWAYOUT
      /* clear the counter */
      inb_p(WDIO);
      /* disable the watchdog */
      inb_p(WDIODIS);
#endif
      rockywd_is_open=0;
   }
   MOD_DEC_USE_COUNT;
   return 0;
}

static ssize_t rockywd_write(struct file *file, const char *buf, size_t count, loff_t *ppos) {
   if (ppos!=&file->f_pos)
     return -ESPIPE;
   if (count) {
      /* clear the counter so we don't go boom */
      inb_p(WDIO);
      return 1;
   }
   return 0;
}

static int rockywd_notify_sys(struct notifier_block *this, unsigned long code,
	void *unused) {
	if(code==SYS_DOWN || code==SYS_HALT) {
	   /* zero the counter */
	   inb_p(WDIO);
	   /* turn the watchdog off */
	   inb_p(WDIODIS);
	}
	return NOTIFY_DONE;
}
 

static struct file_operations rockywd_fops = {
   NULL,
     NULL,
     rockywd_write,
     NULL,
     NULL,
     NULL,
     NULL,
     rockywd_open,
     NULL,
     rockywd_release
};
     
static struct miscdevice rockywd_miscdev = {
   WATCHDOG_MINOR,
     "watchdog",
     &rockywd_fops
};

static struct notifier_block rockywd_notifier = {
   rockywd_notify_sys,
     NULL,
     0
};

#ifdef MODULE
#define rockywd_init init_module

void cleanup_module(void) {
   misc_deregister(&rockywd_miscdev);
   unregister_reboot_notifier(&rockywd_notifier);
   release_region(WDIO, 1);
   release_region(WDIODIS, 2);
}

#endif

__initfunc(int rockywd_init(void)) {
   printk("ROCKY-3702EV watchdog driver $Revision: 1.3 $: <cpbotha@ieee.org>\n");
   misc_register(&rockywd_miscdev);
   request_region(WDIO, 1, "rockywd");
   request_region(WDIODIS, 2, "rockywd");
   register_reboot_notifier(&rockywd_notifier);
   return 0;
}
